#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}

//#define HAVECUDA1
__global__ 
void cuda_sgemm(const float *ad, const float *bd, 
                           const int n, float *cd) {
}

//#define HAVECUDA2
__global__ 
void cuda_sgemm_dblsum(const float *ad, const float *bd, 
                               const int n, float *cd) {
    return;
}

void cpu_sgemm(const float *a, const float *b, 
               const int n, float *c) {

    /* this, of course, is a  
        terrible implementation */
	int i, j, k;
	for (i=0;i<n;i++) {
        for (j=0;j<n;j++) {
            c[i*n + j] = 0.;;
            for (k=0;k<n;k++) {
                c[i*n + j] += a[i*n + k]*b[k*n + j];
            }
		}
	}
	return;
}

int get_options(int argc, char **argv, int *n, int *nb, long int *seed);
void tick(struct timeval *timer);
double tock(struct timeval *timer);
void printmatrix(const char *hdr, const float *a, const int n);

int main(int argc, char **argv) {
	int n=160;
	int nblocks=10;
	int err;
	float *a, *b, *c, *ccuda, *ccuda2;
	float *ad, *bd, *cd;
    float *cdiff, *cdiff2;
	double abserr, abserr2;
	dim3 blocksize, gridsize;
    long int seed;
    int i;
    struct timeval cputimer;
    struct timeval gputimer;
    struct timeval gputimer2;
    double cputime, gputime, gputime2;
    double totflops;

	err = get_options(argc, argv, &n, &nblocks, &seed);
	if (err) return 0;
    
    if ((n % nblocks) != 0) {
        fprintf(stderr,"Error: Number of blocks %d must evenly divide matrix size %d.\n",
                 nblocks, n);
        return -1;
    }

    srand48(seed);

	a = (float *)malloc(n*n*sizeof(float));
	if (!a) return 1;
		
	b = (float *)malloc(n*n*sizeof(float));
	if (!b) {free(a); return 1;}

	c = (float *)malloc(n*n*sizeof(float));
	if (!c) {free(a); free(b); return 1;}

	ccuda = (float *)malloc(n*n*sizeof(float));
	if (!ccuda) {free(c); free(b); free(a); return 1;}

	cdiff = (float *)malloc(n*n*sizeof(float));
	if (!cdiff) {free(ccuda); free(c); free(b); free(a); return 1;}

	ccuda2 = (float *)malloc(n*n*sizeof(float));
	if (!ccuda2) {free(cdiff); free(ccuda); free(c); free(b); free(a); return 1;}

	cdiff2 = (float *)malloc(n*n*sizeof(float));
	if (!cdiff2) {free(ccuda2); free(cdiff); free(ccuda); free(c); free(b); free(a); return 1;}

    for (i=0; i < n*n; i++) {
        a[i] = (float)drand48();
        b[i] = (float)drand48();
    }

    totflops = 2.0*n*n*n;

	/* run CPU code */
    /* no iterations here -- calculation is big enough don't need it */
    tick(&cputimer);
    cpu_sgemm(a, b, n, c);
    cputime = tock(&cputimer);

	/* run GPU code #1 */
    #ifdef HAVECUDA1
        CHK_CUDA( cudaMalloc(&ad, n*n*sizeof(float)) );
        CHK_CUDA( cudaMalloc(&bd, n*n*sizeof(float)) );
        CHK_CUDA( cudaMalloc(&cd, n*n*sizeof(float)) );

        tick(&gputimer);
        CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );
        CHK_CUDA( cudaMemcpy(bd, b, n*n*sizeof(float), cudaMemcpyHostToDevice) );

        blocksize = make_uint3( /*?*/, /*?*/ ,  /*?*/);
        gridsize  = make_uint3( nblocks, nblocks, 1);

        cuda_sgemm<<<gridsize, blocksize>>>(ad, bd, n, cd);

        CHK_CUDA( cudaMemcpy(ccuda, cd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
        gputime = tock(&gputimer);
        
        CHK_CUDA( cudaFree(ad) );
        CHK_CUDA( cudaFree(bd) );
        CHK_CUDA( cudaFree(cd) );
    #endif 

	/* run GPU code #2 */
    #ifdef HAVECUDA2
        CHK_CUDA( cudaMalloc(&ad, n*n*sizeof(float)) );
        CHK_CUDA( cudaMalloc(&bd, n*n*sizeof(float)) );
        CHK_CUDA( cudaMalloc(&cd, n*n*sizeof(float)) );

        tick(&gputimer2);
        CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );
        CHK_CUDA( cudaMemcpy(bd, b, n*n*sizeof(float), cudaMemcpyHostToDevice) );

        blocksize = make_uint3( /*?*/, /*?*/ ,  /*?*/);
        gridsize  = make_uint3( nblocks, nblocks, 1);

        cuda_sgemm_dblsum<<<gridsize, blocksize>>>(ad, bd, n, cd);

        CHK_CUDA( cudaMemcpy(ccuda2, cd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
        gputime2 = tock(&gputimer2);
        
        CHK_CUDA( cudaFree(ad) );
        CHK_CUDA( cudaFree(bd) );
        CHK_CUDA( cudaFree(cd) );
    #endif 

    abserr = 0.;
    abserr2= 0.;
    for (i=0;i<n*n;i++) {
        #ifdef HAVECUDA1
            cdiff[i] = fabs(ccuda[i] - c[i]);
            abserr += cdiff[i];
        #endif
        #ifdef HAVECUDA2 
            cdiff2[i] = fabs(ccuda2[i] - c[i]);
            abserr2 += cdiff2[i];
        #endif
    }
    
    if (n<20) printmatrix("Difference", cdiff, n);
    
    printf("Matrix size = %d, Number of blocks = %d.\n", n, nblocks);
    printf("CPU  time = %lg millisec, GFLOPS=%lf\n", cputime*1000., totflops/(cputime*1.e9));
    #ifdef HAVECUDA1
        printf("GPU  time = %lg millisec, GFLOPS=%lf, diff = %lf.\n", gputime*1000.,totflops/(gputime*1.e9),abserr);
    #endif
    #ifdef HAVECUDA2
        printf("GPU2 time = %lg millisec, GFLOPS=%lf, diff = %lf.\n", gputime2*1000.,totflops/(gputime2*1.e9),abserr2);
    #endif

    free(a);
    free(b);
    free(c);
    free(ccuda);
    free(ccuda2);
    free(cdiff);
    free(cdiff2);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, long int *seed) {

  const struct option long_options[] = {
    {"matsize" , required_argument, 0, 'n'},
    {"nblocks" , required_argument, 0, 'B'},
    {"seed"    , required_argument, 0, 's'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:s:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 50000) {
            fprintf(stderr,"%s: Cannot use number of columns %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'B': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
          } else {
            *nb = tempint;
          }
          break;

      case 's': *seed = atoi(optarg);
          break;

      case 'h':
          puts("Calculates a matrix multiplication on the GPU.");
          puts("Options: ");
          puts("    --matsize=N    (-n N): Set the number of values in y,x.");
          puts("    --nblocks=N    (-B N): Set the number of blocks used in each dimension.");
          puts("    --seed=N       (-s N): Seed the random number generator.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}

void printmatrix(const char *hdr, const float *a, const int n) {
    int i,j;
    printf("%s\n", hdr);

    for (i=0; i<n; i++) {
        for (j=0; j<n; j++) {
            printf("%7.5g ", a[i*n+j]);
        } 
        puts("");
    }
}
